from base64 import b64decode
from io import BytesIO

from PIL import Image

from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class ScreenshotExtractor(Extractor):
    def extract_information(self):
        screenshot = self.page.tab.Page.captureScreenshot(clip={
            'x': 0,
            'y': 0,
            'width': 1920,
            'height': 1080,
            'scale': 1
        }, format='png')
        screenshot = BytesIO(b64decode(screenshot['data']))
        screenshot_pixelized = BytesIO()
        pixelize_screenshot(screenshot, screenshot_pixelized)
        self.result.add_file('screenshot.png', screenshot_pixelized.getvalue())


def pixelize_screenshot(screenshot, screenshot_pixelized, target_width=390, pixelsize=3):
    """
    Thumbnail a screenshot to `target_width` and pixelize it.

    :param screenshot: Screenshot to be thumbnailed in pixelized
    :param screenshot_pixelized: File to which the result should be written
    :param target_width: Width of the final thumbnail
    :param pixelsize: Size of the final pixels
    :return: None
    """
    if target_width % pixelsize != 0:
        raise ValueError("pixelsize must divide target_width")

    img = Image.open(screenshot)
    width, height = img.size
    if height > width:
        img = img.crop((0, 0, width, width))
        height = width
    undersampling_width = target_width // pixelsize
    ratio = width / height
    new_height = int(undersampling_width / ratio)
    img = img.resize((undersampling_width, new_height), Image.BICUBIC)
    img = img.resize((target_width, new_height * pixelsize), Image.NEAREST)
    img.save(screenshot_pixelized, format='png')



"""
This test module does a number of network-based checks to determine web- and mailserver
addresses and the final URL after following any HTTP forwards.
"""
import os
import re
import tempfile
import warnings
from pathlib import Path
import tarfile
from typing import List, Iterable, Tuple
from urllib.parse import urlparse

import requests
import urllib3
from dns import resolver, reversename
from dns.exception import DNSException
from geoip2.database import Reader
from geoip2.errors import AddressNotFoundError

from privacyscanner.scanmodules import ScanModule
from privacyscanner.utils import download_file, copy_to, file_is_outdated


class NetworkScanModule(ScanModule):
    name = 'network'
    dependencies = []
    required_keys = ['site_url']

    def scan_site(self, result, logger, meta):
        scan_site(result, logger, self.options, meta)


# The minimum Jaccard coefficient required for the
# comparison of http and https version of a site
# so that we accept both sites to show the same
# content (if threshold not reached we will report
# that the scanned site is not available via https)
MINIMUM_SIMILARITY = 0.90

# PrivacyScanner's local copy of the GeoIP database
# which takes precedence over copies in other locations
GEOIP_DATABASE_PATH = Path('~/.local/share/privacyscanner/GeoIP/GeoLite2-Country.mmdb').expanduser()
GEOIP_DOWNLOAD_URL = 'https://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz'
GEOIP_MAX_AGE = 3 * 24 * 3600


def scan_site(result, logger, options, meta):
    # determine hostname
    hostname = urlparse(result['site_url']).hostname

    _insert_dns_records(result, logger, options, hostname)
    _insert_geoip(result, logger, options)

    result['reachable'] = True

    # Note: If you ask for a A record, but there is only a CNAME record,
    # the DNS resolver will also return the A record which is referenced
    # by the CNAME, i.e. result['a_records'] won't be empty if there is
    # only a CNAME referring to an A record (see RFC 1034).
    if not result['a_records']:
        result['dns_error'] = True
        result['reachable'] = False
        return

    # Determine final URL. The final URL is the URL that is retrieved
    # after some optional redirects when retrieving the site url.
    final_url = result['site_url']
    try:
        final_url, final_url_content = _retrieve_url(result['site_url'])
    except requests.exceptions.HTTPError as e:
        result['reachable'] = False
        result['http_error'] = str(e)
        final_url_content = e.response.content
    except requests.exceptions.RequestException as e:
        logger.exception('Failed to retrieve URL')
        result['reachable'] = False
        return
    finally:
        result['final_url'] = final_url


    final_url_is_https = final_url.startswith('https://')
    result['final_url_is_https'] = final_url_is_https

    # If our final URL is already HTTPS, we have nothing to do and just
    # set is as HTTPS-URL. Otherwise we fetch the site again with HTTPS
    # and compare it to the HTTP version later on using Jaccard index
    # similarity
    if final_url_is_https:
        result['final_https_url'] = result['final_url']
    else:
        https_url = 'https:/' + result['site_url'].split('/', maxsplit=1)[1]
        try:
            final_https_url, final_https_url_content = _retrieve_url(https_url)
            result['final_https_url'] = final_https_url
        except requests.exceptions.HTTPError as e:
            result['https_error'] = str(e)
            result['final_https_url'] = https_url
            final_https_url_content = e.response.content
        except requests.exceptions.RequestException:
            result['final_https_url'] = None
            return
        else:
            similarity = _jaccard_index(
                final_url_content,
                final_https_url_content)
            minimum_similarity = options.get('minimum_similarity', MINIMUM_SIMILARITY)
            result['same_content_via_https'] = similarity > minimum_similarity


def update_dependencies(options):
    if not file_is_outdated(GEOIP_DATABASE_PATH, GEOIP_MAX_AGE):
        return
    GEOIP_DATABASE_PATH.parent.mkdir(parents=True, exist_ok=True)
    FILES = ['COPYRIGHT.txt', 'LICENSE.txt', 'GeoLite2-Country.mmdb']
    with tempfile.NamedTemporaryFile() as f:
        download_file(GEOIP_DOWNLOAD_URL, f)
        archive = tarfile.open(f.name)
        for member in archive.getmembers():
            base_name = os.path.basename(member.name)
            if base_name in FILES and member.isfile():
                with (GEOIP_DATABASE_PATH.parent / base_name).open('wb') as f:
                    copy_to(archive.extractfile(member), f)


def _insert_dns_records(result, logger, options, hostname):
    # DNS
    # CNAME records
    result['cname_records'] = _cname_lookup(hostname)

    # A records
    result['a_records'] = _a_lookup(hostname)

    # MX records
    result['mx_records'] = _mx_lookup(hostname)
    if hostname.startswith('www.'):
        result['mx_records'] += _mx_lookup(hostname[4:])

    # A records for MX
    result['mx_a_records'] = [(pref, _a_lookup(mx)) for pref, mx in result['mx_records']]

    # Reverse A
    result['a_records_reverse'] = [_reverse_lookup(a) for a in result['a_records']]

    # Reverse A for MX
    result['mx_a_records_reverse'] = [
        (pref, [_reverse_lookup(a) for a in mx_a])
        for pref, mx_a in result['mx_a_records']]


def _insert_geoip(result, logger, options):
    # GeoIP
    reader = Reader(_get_geoip_database(options))

    result['a_locations'], result['a_locations_unknown'] = _get_countries(
        result['a_records'], reader)
    result['mx_locations'], result['mx_locations_unknown'] = _get_countries(
        (ip for mx_a_records in result['mx_a_records']
         for ip in mx_a_records[1]), reader)


def _retrieve_url(url):
    headers = {
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:53.0) Gecko/20100101 Firefox/53.0'
    }

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', urllib3.exceptions.InsecureRequestWarning)
        r = requests.get(url, headers=headers, verify=False)
    r.raise_for_status()
    return r.url, r.content


def _a_lookup(name: str) -> List[str]:
    try:
        return [e.address for e in resolver.query(name, 'A')]
    except DNSException:
        return []


def _cname_lookup(name: str) -> List[str]:
    try:
        return [e.to_text()[:-1].lower() for e in resolver.query(name, 'CNAME')]
    except DNSException:
        return []


def _mx_lookup(name: str) -> Iterable[Tuple[str, str]]:
    try:
        return sorted(((e.preference, e.exchange.to_text()[:-1].lower())
                       for e in resolver.query(name, 'MX')), key=lambda v: v[0])
    except DNSException:
        return []


def _reverse_lookup(ip: str) -> List[str]:
    try:
        address = reversename.from_address(ip).to_text()
        return [rev.to_text()[:-1].lower()
                for rev in resolver.query(address, 'PTR')]
    except DNSException:
        return []


def _get_countries(addresses: Iterable[str], reader: Reader) -> Tuple[List[str], List[str]]:
    locations = set()
    unresolved_ips = set()
    for ip in addresses:
        try:
            geoip_result = reader.country(ip)
            this_result = geoip_result.country.name
            if not this_result:
                this_result = geoip_result.continent.name
            if not this_result:
                raise AddressNotFoundError
            locations.add(this_result)
        except AddressNotFoundError:
            unresolved_ips.add(ip)
            continue
    return list(locations), list(unresolved_ips)


def _jaccard_index(a: bytes, b: bytes) -> float:
    """Calculate the jaccard similarity of a and b."""
    pattern = re.compile(rb'[ \n]')
    # remove tokens containing / to prevent wrong classifications for
    # absolute paths
    a = {token for token in pattern.split(a) if b'/' not in token}
    b = {token for token in pattern.split(b) if b'/' not in token}
    intersection = a.intersection(b)
    union = a.union(b)
    return len(intersection) / len(union)


def _get_geoip_database(options):
    try:
        return options['country_database_path']
    except KeyError:
        possible_paths = [
            GEOIP_DATABASE_PATH,
            Path('~/.local/share/GeoIP/GeoLite2-Country.mmdb').expanduser(),
            Path('/var/lib/GeoIP/GeoLite2-Country.mmdb'),
            Path('/usr/local/var/GeoIP/GeoLite2-Country.mmdb'),
            Path('/usr/local/share/GeoIP/GeoLite2-Country.mmdb'),
            Path('/usr/share/GeoIP/GeoLite2-Country.mmdb'),
        ]
        for p in possible_paths:
            if p.exists():
                return str(p)
    raise RuntimeError('Could not find GeoIP database. '
                       'Please specify country_database_path option.')



import argparse
import hashlib
import json
import logging
import os
import pprint
import string
import sys
import tempfile
import time
from collections import namedtuple
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from urllib.parse import urlparse

from toposort import toposort, toposort_flatten

from privacyscanner.filehandlers import DirectoryFileHandler
from privacyscanner.raven import has_raven, raven
from privacyscanner.result import Result
from privacyscanner.scanmeta import ScanMeta
from privacyscanner.scanmodules import load_modules
from privacyscanner import defaultconfig
from privacyscanner.loghandlers import ScanFileHandler, ScanStreamHandler
from privacyscanner.exceptions import RescheduleLater, RetryScan

CONFIG_LOCATIONS = [
    Path('~/.config/privacyscanner/config.py').expanduser(),
    Path('/etc/privacyscanner/config.py')
]


class CommandError(Exception):
    pass


QueueEntry = namedtuple('QueueEntry', ['scan_module_name', 'num_try', 'not_before'])


def load_config(config_file):
    config = deepcopy(defaultconfig.__dict__)
    if config_file is None:
        for filename in CONFIG_LOCATIONS:
            if filename.is_file():
                config_file = filename
                break
        else:
            return config
    config_file = Path(config_file)
    try:
        with config_file.open() as f:
            code = compile(f.read(), config_file.name, 'exec')
            exec(code, {}, config)
            return config
    except IOError as e:
        raise CommandError('Could not open config: {}'.format(e)) from e
    except Exception as e:
        raise CommandError('Could not parse config: {}: {}'.format(e.__class__.__name__, e)) from e


def slugify(somestr):
    allowed_chars = string.ascii_lowercase + string.digits + '.-'
    return ''.join(x for x in somestr.lower() if x in allowed_chars)


def run_workers(args):
    from .worker import WorkerMaster

    config = load_config(args.config)
    raven_client = None
    if has_raven and config['RAVEN_DSN']:
        raven_client = raven.Client(config['RAVEN_DSN'])
    master = WorkerMaster(config['QUEUE_DB_DSN'], config['SCAN_MODULES'],
                          config['SCAN_MODULE_OPTIONS'], config['MAX_TRIES'],
                          config['NUM_WORKERS'], config['MAX_EXECUTIONS'],
                          config['MAX_EXECUTION_TIMES'], config['RAVEN_DSN'])
    try:
        master.start()
    except Exception:
        raven_client.captureException()


def scan_site(args):
    config = load_config(args.config)

    site_parsed = urlparse(args.site)
    if site_parsed.scheme not in ('http', 'https'):
        raise CommandError('Invalid site: {}'.format(args.site))

    results_dir = args.results
    if results_dir is None:
        results_dir = slugify(site_parsed.netloc) + '_'
        results_dir += hashlib.sha512(args.site.encode()).hexdigest()[:10]
    results_dir = Path(results_dir).resolve()
    try:
        results_dir.mkdir(exist_ok=True)
    except IOError as e:
        raise CommandError('Could not create results directory: {}'.format(e)) from e

    result_file = results_dir / 'results.json'
    result_json = {'site_url': args.site}
    if args.import_results:
        try:
            with open(args.import_results) as f:
                import_json = json.load(f)
        except IOError as e:
            raise CommandError('Could not open result JSON: {}'.format(e)) from e
        except ValueError as e:
            raise CommandError('Could not parse result JSON: {}'.format(e)) from e
        else:
            result_json.update(import_json)
    try:
        with result_file.open('w') as f:
            json.dump(result_json, f, indent=2)
            f.write('\n')
    except IOError as e:
        raise CommandError('Could not write result JSON: {}'.format(e)) from e

    scan_modules = load_modules(config['SCAN_MODULES'],
                                       config['SCAN_MODULE_OPTIONS'])
    scan_module_names = args.scan_modules

    if scan_module_names is None:
        scan_module_names = scan_modules.keys()

    # Order scan_module_names by dependency topologically
    dependencies = {}
    for scan_module_name in scan_module_names:
        mod = scan_modules[scan_module_name]
        dependencies[mod.name] = set(mod.dependencies)
    scan_module_names = toposort_flatten(dependencies)

    if args.skip_dependencies:
        scan_module_names = [
            scan_module_name
            for scan_module_name in scan_module_names
            if scan_module_name in args.scan_modules
        ]

    has_error = False
    result = Result(result_json, DirectoryFileHandler(results_dir))
    stream_handler = ScanStreamHandler()
    logs_dir = results_dir / 'logs'
    logs_dir.mkdir(exist_ok=True)
    scan_queue = [QueueEntry(mod_name, 0, None) for mod_name in scan_module_names]
    scan_queue.reverse()
    while scan_queue:
        scan_module_name, num_try, not_before = scan_queue.pop()
        if not_before is not None:
            # noinspection PyTypeChecker
            while datetime.utcnow() < not_before:
                time.sleep(0.5)
        mod = scan_modules[scan_module_name]
        num_try += 1
        log_filename = (logs_dir / (mod.name + '.log'))
        file_handler = ScanFileHandler(str(log_filename))
        logger = logging.Logger(mod.name)
        logger.addHandler(stream_handler)
        logger.addHandler(file_handler)
        scan_meta = ScanMeta(worker_id=0, num_tries=num_try)
        with tempfile.TemporaryDirectory() as temp_dir:
            old_cwd = os.getcwd()
            os.chdir(temp_dir)
            logger.info('Starting %s', mod.name)
            try:
                mod.scan_site(result, logger, scan_meta)
            except RetryScan:
                if num_try <= config['MAX_TRIES']:
                    scan_queue.append(QueueEntry(scan_module_name, num_try, not_before))
                    logger.info('Scan module `%s` will be retried', mod.name)
                else:
                    has_error = True
            except RescheduleLater as e:
                scan_queue.append(QueueEntry(scan_module_name, num_try, e.not_before))
            except Exception:
                if num_try <= config['MAX_TRIES']:
                    scan_queue.append(QueueEntry(scan_module_name, num_try, not_before))
                has_error = True
                logger.exception('Scan module `%s` failed.', mod.name)
            finally:
                os.chdir(old_cwd)
                with result_file.open('w') as f:
                    json.dump(result.get_results(), f, indent=2, sort_keys=True)
                    f.write('\n')
            logger.info('Finished %s', mod.name)
    pprint.pprint(result.get_results())
    if has_error:
        sys.exit(1)


def update_dependencies(args):
    config = load_config(args.config)
    scan_modules = load_modules(config['SCAN_MODULES'],
                                config['SCAN_MODULE_OPTIONS'])
    updated = []
    for scan_module in scan_modules.values():
        print(scan_module.name)
        if hasattr(scan_module, 'update_dependencies'):
            scan_module.update_dependencies()
            updated.append(scan_module.name)
    if updated:
        print(' '.join(updated))
    else:
        print('Nothing to update.')


def print_master_config(args):
    config = load_config(args.config)
    scan_modules = load_modules(config['SCAN_MODULES'])
    dependencies = {}
    for scan_module in scan_modules.values():
        dependencies[scan_module.name] = set(scan_module.dependencies)
    modules_topology = {}
    for index, module_list in enumerate(toposort(dependencies)):
        for module_name in module_list:
            modules_topology[module_name] = index
    output = '# Scan modules with topological dependency order index.\n'
    output += '# Run the following to obtain this configuration value:\n'
    output += '# privacyscanner print_master_config --config yourconfig.py\n'
    output += 'SCAN_MODULES = {}'.format(pprint.pformat(modules_topology, indent=4))
    print(output)


def main():
    parser = argparse.ArgumentParser(description='Scan sites for privacy.')
    subparsers = parser.add_subparsers(dest='command')

    parser_run_workers = subparsers.add_parser('run_workers')
    parser_run_workers.add_argument('-c', '--config', help='Configuration_file')
    parser_run_workers.set_defaults(func=run_workers)

    parser_scan = subparsers.add_parser('scan')
    parser_scan.add_argument('site', help='Site to scan')
    parser_scan.add_argument('-c', '--config', help='Configuration_file')
    parser_scan.add_argument('--results', help='Directory to store results')
    parser_scan.add_argument('--import-results', dest='import_results',
                             help='Import JSON results from a file before scanning')
    parser_scan.add_argument('-m', '--scan-modules', dest='scan_modules',
                             type=lambda scans: [x.strip() for x in scans.split(',')],
                             help='Comma separated list of scan modules')
    parser_scan.add_argument('--skip-dependencies', action='store_true',
                             help='Do not run dependencies that are not explicitly '
                                  'specified using --scans')
    parser_scan.add_argument('--print', dest='print_result', action='store_true')
    parser_scan.set_defaults(func=scan_site)

    parser_print_master_config = subparsers.add_parser('print_master_config')
    parser_print_master_config.add_argument('-c', '--config', help='Configuration_file')
    parser_print_master_config.set_defaults(func=print_master_config)

    parser_run_workers = subparsers.add_parser('update_dependencies')
    parser_run_workers.add_argument('--config', help='Configuration_file')
    parser_run_workers.set_defaults(func=update_dependencies)

    args = parser.parse_args()
    if args.command is None:
        parser.error('No arguments')
    if args.command == 'scan' and args.skip_dependencies and not args.scan_modules:
        parser.error('--skip-dependencies can only be set when using --scans')
    try:
        args.func(args)
    except CommandError as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()



# Python 3.5 compatibility
try:
    ModuleNotFoundError
except NameError:
    ModuleNotFoundError = ImportError

try:
    import raven
    has_raven = True
except ModuleNotFoundError:
    raven = None
    has_raven = False



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import camelcase_to_underscore


class InsecureContentExtractor(Extractor):
    def extract_information(self):
        entry = self.page.security_state_log[-1]
        insecure_content = {}
        # See https://chromedevtools.github.io/devtools-protocol/tot/Security#type-InsecureContentStatus
        properties = [
            # True if the page was loaded over HTTPS and ran mixed
            # (HTTP) content such as scripts.
            'ranMixedContent',
            # True if the page was loaded over HTTPS and displayed
            # mixed (HTTP) content such as images.
            'displayedMixedContent',
            # True if the page was loaded over HTTPS and contained a
            # form targeting an insecure url.
            'containedMixedForm',
            # True if the page was loaded over HTTPS without
            # certificate errors, and ran content such as scripts that
            # were loaded with certificate errors.
            'ranContentWithCertErrors',
            # True if the page was loaded over HTTPS without
            # certificate errors, and displayed content such as images
            # that were loaded with certificate errors.
            'displayedContentWithCertErrors'
        ]
        status = entry['insecureContentStatus']
        for key, value in status.items():
            if key not in properties:
                continue
            insecure_content[camelcase_to_underscore(key)] = value
        insecure_content['has_mixed_content'] = (status['ranMixedContent'] or
                                                 status['displayedMixedContent'])
        insecure_content['has_cert_errors'] = (status['ranContentWithCertErrors'] or
                                               status['displayedContentWithCertErrors'])
        self.result['insecure_content'] = insecure_content



from privacyscanner.scanmodules import ScanModule
from privacyscanner.scanmodules.chromedevtools.chromescan import ChromeScan
from privacyscanner.scanmodules.chromedevtools.extractors import FinalUrlExtractor, \
    GoogleAnalyticsExtractor, CookiesExtractor, RequestsExtractor,  RedirectChainExtractor, \
    TLSDetailsExtractor, CertificateExtractor, ThirdPartyExtractor, InsecureContentExtractor, \
    FailedRequestsExtractor, SecurityHeadersExtractor, TrackerDetectExtractor, \
    CookieStatsExtractor, JavaScriptLibsExtractor, ScreenshotExtractor, ImprintExtractor
from privacyscanner.scanmodules.chromedevtools.utils import TLDEXTRACT_CACHE_FILE, parse_domain
from privacyscanner.utils import file_is_outdated


EXTRACTOR_CLASSES = [FinalUrlExtractor, RedirectChainExtractor, GoogleAnalyticsExtractor,
                     CookiesExtractor, RequestsExtractor, TLSDetailsExtractor,
                     CertificateExtractor, ThirdPartyExtractor, InsecureContentExtractor,
                     FailedRequestsExtractor, SecurityHeadersExtractor, TrackerDetectExtractor,
                     CookieStatsExtractor, JavaScriptLibsExtractor, ScreenshotExtractor,
                     ImprintExtractor]


class ChromeDevtoolsScanModule(ScanModule):
    name = 'chromedevtools'
    dependencies = []
    required_keys = ['site_url']

    def scan_site(self, result, logger, meta):
        chrome_scan = ChromeScan(EXTRACTOR_CLASSES)
        debugging_port = self.options.get('start_port', 9222) + meta.worker_id
        chrome_scan.scan(result, logger, self.options, meta, debugging_port)

    def update_dependencies(self):
        max_age = 14 * 24 * 3600
        TLDEXTRACT_CACHE_FILE.parent.mkdir(parents=True, exist_ok=True)
        if file_is_outdated(TLDEXTRACT_CACHE_FILE, max_age):
            parse_domain.update(fetch_now=True)
        for extractor_class in EXTRACTOR_CLASSES:
            if hasattr(extractor_class, 'update_dependencies'):
                extractor_class.update_dependencies(self.options)



import importlib


class ModuleLoadError(Exception):
    pass


class ScanModule:
    def __init__(self, options):
        self.options = options

    def scan_site(self, result, logger, meta):
        raise NotImplemented

    def update_dependencies(self):
        pass


def load_modules(module_list, module_options):
    scan_modules = {}
    for module_class in module_list:
        module_name, class_name = module_class.rsplit('.', 1)
        try:
            mod = importlib.import_module(module_name)
            class_obj = getattr(mod, class_name)
        except (ImportError, AttributeError) as e:
            raise ModuleLoadError('Could not load module {}.'.format(module_class)) from e
        for attr_name in ('name', 'scan_site', 'dependencies', 'required_keys'):
            if not hasattr(class_obj, attr_name):
                raise ModuleLoadError('Module {} has no attribute `{}`.'.format(
                    module_class, attr_name
                ))
        options = module_options.get(class_obj.name, {})
        scan_modules[class_obj.name] = class_obj(options.get(class_obj.name, {}))
    return scan_modules



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import javascript_evaluate, JavaScriptError


LIBRARY_JS = """
(function() {
    return {
        'jQuery': typeof(jQuery) !== 'undefined' ? jQuery.fn.jquery : null,
        'React': typeof(React) !== 'undefined' ? React.version : null,
        'AngularJS': typeof(angular) !== 'undefined' ? angular.version.full : null
    };
})()
"""


class JavaScriptLibsExtractor(Extractor):
    def extract_information(self):
        versions = {
            'jQuery': None,
            'React': None,
            'AngularJS': None
        }
        try:
            versions.update(javascript_evaluate(self.page.tab, LIBRARY_JS))
        except JavaScriptError:
            pass
        self.result['javascript_libraries'] = versions



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class SecurityHeadersExtractor(Extractor):
    def extract_information(self):
        response = self.page.final_response
        if response is None:
            self.logger.error('Could not find response for final url')
            return
        headers = response['headers_lower']

        header_names = ['Referrer-Policy', 'X-Content-Type-Options',
                        'X-Frame-Options', 'Expect-CT',
                        'Access-Control-Allow-Origin']
        security_headers = {}
        for header_name in header_names:
            security_headers[header_name] = self._get_header(headers, header_name)

        hsts_value = None
        if 'strict-transport-security' in headers:
            hsts_value = self._parse_hsts(headers['strict-transport-security'])
        security_headers['Strict-Transport-Security'] = hsts_value

        csp_value = None
        if 'content-security-policy' in headers:
            csp_value = self._parse_csp(headers['content-security-policy'])
        security_headers['Content-Security-Policy'] = csp_value

        xss_protection = None
        if 'x-xss-protection' in headers:
            xss_protection = self._parse_xss_protection(headers['x-xss-protection'])
        security_headers['X-XSS-Protection'] = xss_protection

        self.result['security_headers'] = security_headers

    @staticmethod
    def _parse_csp(header_value):
        csp = {}
        parts = [part.strip() for part in header_value.split(';')]
        for part in parts:
            if not part:
                continue
            values = part.split()
            key = values[0]
            values = values[1:]
            csp[key.lower()] = values
        csp['header_value'] = header_value
        return csp

    @staticmethod
    def _parse_hsts(header_value):
        parts = [part.strip() for part in header_value.split(';')]
        max_age = None
        for part in parts:
            if part.startswith('max-age='):
                max_age = part.split('=', 1)[1]
                try:
                    max_age = int(max_age)
                except ValueError:
                    pass
                break
        return {
            'header_value': header_value,
            'includeSubDomains': 'includeSubDomains' in parts,
            'preload': 'preload' in parts,
            'max-age': max_age
        }

    @staticmethod
    def _parse_xss_protection(header_value):
        mode = None
        is_active = None
        if ';' in header_value:
            is_active, mode_str = header_value.split(';', 1)
            is_active = is_active.strip() == '1'
            if mode_str.strip().startswith('mode='):
                mode = mode_str.split('=', 1)[1].strip()
        return {
            'header_value': header_value,
            'is_active': is_active,
            'mode': mode
        }

    @staticmethod
    def _get_header(headers, header_name):
        header_name = header_name.lower()
        if header_name in headers:
            value = headers[header_name]
            # Chrome will separate multiple headers with a newline,
            # however, RFC 2616 says that they should be interpreted
            # with a comma in between. See RFC2616 Sect. 4.2:
            # https://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2
            value = value.replace('\n', ',')
            return value
        return None



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class RequestsExtractor(Extractor):
    def extract_information(self):
        requests = []

        for request in self.page.request_log:
            if request['url'].startswith('data:'):
                continue
            response = self.page.get_final_response_by_id(request['requestId'])
            request_dict = {
                'url': request['url'],
                'sets_cookie': self._get_sets_cookie(response),
                'mime_type': response['mimeType'] if response else None,
                'status_code': response['status'] if response else None,
                'status_text': response['statusText'] if response else None
            }
            # is_thirdparty is only availavle if the thirdparties mixin is enabled
            if 'is_thirdparty' in request:
                request_dict['is_thirdparty'] = request['is_thirdparty']
            # is_tracker is only available if the trackerdetect mixin is enabled
            if 'is_tracker' in request:
                request_dict['is_tracker'] = request['is_tracker']
            # Add headers if requested
            # To enable this option, set SCAN_MODULE_OPTIONS in your config file to
            # {'chromedevtools': {'RequestsExtractor.save_headers': True}}
            # (or change it in a similar way)
            if self.options.get('RequestsExtractor.save_headers', False):
                request_dict['request_headers'] = request["headers"]
                request_dict['response_headers'] = response["headers"]
            requests.append(request_dict)
        self.result['requests'] = requests

    @staticmethod
    def _get_sets_cookie(response):
        if response is None:
            return False
        return 'set-cookie' in response['headers_lower']



import dns.resolver

from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import parse_domain


class FailedRequestsExtractor(Extractor):
    def extract_information(self):
        requests_lookup = {request['requestId']: request for request in self.page.request_log}
        failed_requests = []
        for failed_request in self.page.failed_request_log:
            error_text = failed_request['errorText']
            valid_errors = ('net::ERR_CACHE_MISS', 'net::ERR_ABORTED')
            if any(error in error_text for error in valid_errors):
                # Requests that were aborted by the site (e.g. a XHR
                # request that was canceled) and cache misses are
                # not considered failed.
                continue
            extra = None
            try:
                request = requests_lookup[failed_request['requestId']]
            except KeyError:
                self.logger.error('Could not find request: {}'.format(failed_request))
                continue
            if 'net::ERR_NAME_NOT_RESOLVED' in error_text:
                error_type = 'dns-not-resolved'
                # We could not resolve the IP address of this host. One
                # reason might be, that the domain is not registered.
                # To check whether this is the case, we check for the
                # absence of a SOA record for the domain itself, i.e.,
                # not the netloc of the URL. Unregistered domains
                # should have no SOA entry, while registered should.
                domain = parse_domain(request['url']).registered_domain
                try:
                    dns.resolver.query(domain, 'SOA')
                    domain_registered = True
                # If we have a timeout, we better don't say anything about
                # this domain rather than giving a wrong impressing wether
                # the domain is registered or net
                except dns.resolver.Timeout:
                    domain_registered = None
                # Nameservers behave weird, if the domain is not registered.
                # Some send NXDOMAIN as expected, others prefer to give an
                # answer but do not include a SOA entry in the response.
                # Sometimes all nameservers do not like to answer if the
                # domain is not registered. It is a real mess.
                except (dns.resolver.NXDOMAIN, dns.resolver.NoNameservers,
                        dns.resolver.NoAnswer):
                    domain_registered = False
                extra = {'domain_registered': domain_registered}
            elif 'net::ERR_UNKNOWN_URL_SCHEME' in error_text:
                error_type = 'unknown-url-scheme'
            else:
                error_type = 'unknown'
            error = {
                'url': request['url'],
                'error_type': error_type,
            }
            if extra is not None:
                error.update(extra)
            if error_type == 'unknown':
                error['error_text'] = error_text
            failed_requests.append(error)
        self.result['failed_requests'] = failed_requests



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class RedirectChainExtractor(Extractor):
    def extract_information(self):
        request_id = self.page.request_log[0]['requestId']
        response_chain = self.page.get_response_chain_by_id(request_id)
        self.result['redirect_chain'] = [response['url'] for response in response_chain]



import logging


class WorkerWritePipeHandler(logging.Handler):
    def __init__(self, pid, write_pipe, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pid = pid
        self.write_pipe = write_pipe
        fmt = '%(message)s (%(filename)s:%(lineno)d)'
        self.setFormatter(logging.Formatter(fmt))

    def emit(self, record):
        message = self.format(record)
        self.write_pipe.send((self.pid, 'log', (record.created, record.levelno, message)))


class ScanFileHandler(logging.FileHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fmt = '%(asctime)s: [%(levelname)s] %(message)s (%(filename)s:%(lineno)d)'
        self.setFormatter(logging.Formatter(fmt))


class ScanStreamHandler(logging.StreamHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fmt = '%(name)s> %(asctime)s: [%(levelname)s] %(message)s (%(filename)s:%(lineno)d)'
        self.setFormatter(logging.Formatter(fmt))



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import camelcase_to_underscore


class TLSDetailsExtractor(Extractor):
    def extract_information(self):
        self.result['tls'] = {'has_tls': None}

        response = self.page.final_response
        if response is None:
            self.logger.error('Could not get response for final_url')
            return

        if 'securityDetails' not in response:
            self.result['tls']['has_tls'] = False
            return

        details = {}
        # See https://chromedevtools.github.io/devtools-protocol/tot/Network#type-SecurityDetails
        properties = [
            # Protocol name (e.g. "TLS 1.2" or "QUIC")
            'protocol',
            # Key Exchange used by the connection, or the empty string
            # if not applicable.
            'keyExchange',
            # (EC)DH group used by the connection, if applicable.
            'keyExchangeGroup',
            # Cipher name.
            'cipher',
            # TLS MAC. Note that AEAD ciphers do not have separate MACs.
            'mac',
            # TODO: The following might be also interesting
            # signedCertificateTimestampList
            # certificateTransparencyCompliance
        ]

        for key, value in response['securityDetails'].items():
            if key not in properties:
                continue
            details[camelcase_to_underscore(key)] = value

        self.result['tls']['has_tls'] = True
        self.result['tls'].update(details)



import time
from urllib.request import Request, urlopen


FAKE_UA = 'Mozilla/5.0 (X11; Linux x86_64; rv:61.0) Gecko/20100101 Firefox/61.0'


def download_file(url, fileobj):
    request = Request(url)
    request.add_header('User-Agent', FAKE_UA)
    response = urlopen(request)
    copy_to(response, fileobj)
    fileobj.flush()


def copy_to(src, dest):
    while True:
        data = src.read(8192)
        if not data:
            break
        dest.write(data)


def file_is_outdated(path, max_age):
    try:
        return path.stat().st_mtime + max_age < time.time()
    except FileNotFoundError:
        return True



class ScanMeta:
    def __init__(self, worker_id, num_tries):
        self.worker_id = worker_id
        self.num_tries = num_tries

    @property
    def is_first_try(self):
        return self.num_tries == 1



from base64 import b64decode
from binascii import hexlify
from datetime import datetime

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.dsa import DSAPublicKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import load_der_x509_certificate

from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class CertificateExtractor(Extractor):
    def extract_information(self):
        explanations = self.page.security_state_log[-1]['explanations']
        cert_chain = None
        for explanation in explanations:
            if 'certificate' in explanation:
                cert_chain = explanation['certificate']
                break
        if cert_chain:
            if self.result['tls']['has_tls'] is None:
                self.result['tls']['has_tls'] = True
            self.result['tls']['certificate'] = self._get_certificate_info(cert_chain)

    def _get_certificate_info(self, cert_chain):
        # See https://cryptography.io/en/latest/x509/reference/#cryptography.x509.Certificate
        cert = load_der_x509_certificate(b64decode(cert_chain[0]), backend=default_backend())
        public_key = cert.public_key()
        key_info = {'size': public_key.key_size}
        if isinstance(public_key, RSAPublicKey):
            key_type = 'RSA'
        elif isinstance(public_key, DSAPublicKey):
            key_type = 'DSA'
        elif isinstance(public_key, EllipticCurvePublicKey):
            key_type = 'ECC'
            key_info['curve'] = public_key.curve.name
        else:
            raise ValueError('Invalid key type.')
        key_info['type'] = key_type
        return {
            'version': cert.version.name,
            'fingerprint_sha256': hexlify(cert.fingerprint(hashes.SHA256())).decode(),
            'serial_number': cert.serial_number,
            'not_valid_before': cert.not_valid_before.timestamp(),
            'not_valid_after': cert.not_valid_after.timestamp(),
            'issuer':  {attr.oid._name: attr.value for attr in cert.issuer},
            'subject': {attr.oid._name: attr.value for attr in cert.subject},
            'key': key_info,
            'is_expired': datetime.now() > cert.not_valid_after
        }
        # TODO: Add extensions
